#!/bin/bash

src_root=".."

LOG="configure.log"

scripts="tcl php python lua ruby ghli perl stutter cli guile"

. lib/fstools.lib
. lib/compile.lib
. lib/db.lib
. lib/subst.lib
. lib/test-awk.sh


# Databases:
#		DB_fstools        substitution database for fs tools (key: tool)
#		DB_includes       script module cc include flags (key: language)
#		DB_libs           script module cc lib flags (key: language)
#		DB_osdep          os dependent settings
#		DB_install_paths  installation paths

DB_libs=""
DB_includes=""

#----------------------------- Config part -------------------------------

. configure.defaults
. data/configure_params.tbl

function process_params() {
for n in "$@"
do
	case "$n" in
		--help)
			./help.sh
			exit
			;;
		--include=*)
			value=`awk 'BEGIN { FS="=";  $0="'$n'"; print $2;}'`
			echo "including $value"
			echo "[cmdline] including $value" >> $LOG
			. $value
			;;
		--with*=*)
			name=`awk 'BEGIN { FS="="; $0="'$n'"; gsub("^--", "", $1); gsub("-", "_", $1); print $1;}'`
			value=`awk 'BEGIN { FS="=";  $0="'$n'"; print $2;}'`
			echo "[cmdline] --with: '$name' = '$value'" >> $LOG
			export $name=$value
			;;
		--without*)
			name=`awk 'BEGIN { FS="="; $0="'$n'"; gsub("^--without", "with", $1); gsub("-", "_", $1); print $1;}'`
			echo "[cmdline] --without: '$name'" >> $LOG
			export $name=""
			;;
		*)
			name=`awk 'BEGIN { FS="="; $0="'$n'"; gsub("^--", "", $1); gsub("-", "_", $1); print $1;}'`
			value=`awk 'BEGIN { FS="=";  $0="'$n'"; print $2;}'`
			real_name=`db_get "$name" configure_params`
			if test -z "$real_name"
			then
				echo "Uknown parameter: $n"
				echo "[cmdline] Uknown parameter: $n" >> $LOG
				exit 1
			else
				export $real_name=$value
			fi
			;;
	esac
done
}

#----------------------------- Paths -----------------------------------------
function set_install_paths() {
	# if install paths are relative, insert install_prefix
	for n in with_install_path_doc with_install_path_man with_install_path_bin with_install_path_lib with_install_path_include
	do
		case ${!n} in
			/*) exrpot $n;;
			*) export $n="${with_install_prefix}/${!n}" ;;
		esac
	done

	# These paths can't be changed from settings yet
	export with_install_path_lib_gpmi=$with_install_path_lib/gpmi
	export with_install_path_include_gpmi=$with_install_path_include/gpmi
	export with_install_path_pkg_lib=$with_install_path_lib/gpmi/gpmi_pkg
	export with_install_path_mod_lib=$with_install_path_lib/gpmi/modules
	export with_install_path_pkg_include=$with_install_path_include/gpmi_pkg
	export with_install_path_user_pkg_include=$with_install_path_user_prefix/gpmi_pkg
	export with_install_path_user_pkg_lib=$with_install_path_user_prefix/gpmi_pkg
	export with_install_path_user_mod_lib=$with_install_path_user_prefix/modules

	# Add all paths to DB_install_paths
	for n in install_path_doc install_path_man install_path_bin install_path_lib install_path_include install_path_lib_gpmi install_path_include_gpmi install_path_pkg_lib install_path_mod_lib install_path_pkg_include install_prefix install_path_user_pkg_include install_path_user_pkg_lib install_path_user_mod_lib
	do
		m=with_$n
		db_add $n "${!m}" DB_install_paths
	done
}

#----------------------------- Detect os tools -------------------------------
function os_detect() {
	if test -z "$with_target"
	then
		if ! test -z "$with_target_dir"
		then
			export with_target="$with_target_dir-"
		else
			export with_target=""
		fi
	fi
	if test -z "$with_host"
	then
		if ! test -z "$with_host_dir"
		then
			export with_host="$with_host_dir-"
		else
			export with_host=""
		fi
	fi

	# If host and target don't match, we can safely assume it is a cross-compile
	if ! test "$with_host" = "$with_target"
	then
		with_cross="yes"
	fi

	if test -z "$with_os"
	then
		echo -n "Autodetecting os... "
		echo "[OS] Autodetecting os... " >> $LOG
		with_os=`lib/os-detect.sh`
		if test -z "$with_os"
		then
			echo "FAILED."
			echo ""
			echo "Critical: can't autodetect os."
			echo "[OS] Critical: can't autodetect os." >> $LOG
			exit
		fi
		echo "found $with_os"
		echo "[OS] found '$with_os'" >> $LOG

	else
		echo "Using user supplied os param: $with_os"
		echo "[OS] Using user supplied os param: $with_os" >> $LOG
	fi
	echo -n "Configuring platform-dependent settings..."
	echo "[OS] Configuring platform-dependent settings..." >> $LOG
	if ! test -f "os/${with_os}.sh"
	then
		echo "FAILED"
		echo ""
		echo "Critical: os/${with_os}.sh is not found, so I don't know how to handle operating system ${with_os}."
		echo "[OS] Critical: os/${with_os}.sh is not found, so I don't know how to handle operating system ${with_os}." >> $LOG
		exit
	fi

	echo "[OS] including os/${with_os}.sh" >> $LOG
	. os/${with_os}.sh

	if test "$?" != "0"
	then
		echo "FAILED"
		echo ""
		echo "Critical: os/${with_os}.sh exited with error status $?."
		echo "[OS] Critical: os/${with_os}.sh exited with error status $?." >> $LOG
		exit
	fi
	echo "Done."

	echo -n "Setting up platform dependency database..."
	echo "[OS] Setting up platform dependency database..." >> $LOG

	#set up os dependent DB
	db_add sys_cflags "$with_sys_cflags" DB_osdep
	db_add sys_ldflags "$with_sys_ldfags" DB_osdep
	db_add sys_ldflags_static "$with_sys_ldflags_static" DB_osdep
	db_add sys_ldflags_dynamic "$with_sys_ldflags_dynamic" DB_osdep
	db_add sys_libext_static "$with_sys_libext_static" DB_osdep
	db_add sys_libext_dynamic "$with_sys_libext_dynamic" DB_osdep
	db_add sys_libext_exe "$with_sys_libext_exe" DB_osdep
	db_add sys_cc "$with_cc" DB_osdep
	db_add sys_cc_host "$with_cc_host" DB_osdep
	db_add sys_ld "$with_ld" DB_osdep
	db_add sys_strip "$with_strip" DB_osdep
	db_add sys_ldl "$with_sys_ldl" DB_osdep
	db_add sys_fpic "$with_sys_fpic" DB_osdep
	db_add sys_rdynamic "$with_sys_rdynamic" DB_osdep
	db_add sys_archive_objs "$with_sys_archive_objs" DB_osdep
	db_add with_os "$with_os" DB_osdep
	db_add target_dir "$with_target_dir" DB_osdep
	db_add target "$with_target" DB_osdep
	db_add host_dir "$with_host_dir" DB_osdep
	db_add host "$with_host" DB_osdep

	# Export some variables that might be needed by other detect scripts
	export with_sys_ldl
	export with_sys_fpic
	export with_sys_ld
	export with_sys_strip
	export with_cc
	export with_cc_host
	export with_os
	export with_sys_libext_exe
	export with_sys_cflags
	export with_sys_ldflags
	export with_sys_emulator

	echo "Done."
	echo "[OS] Done." >> $LOG
}

#----------------------------- Detect fs tools -------------------------------
function fs_test() {
	echo -n "fstools: detecting $1... "
	echo "[FS] fstools: detecting $1... " >> $LOG

	if test -z "${!2}"
	then
		tester=$3
		name=$2
		array=$1
		shift 3
		for n in "$1" "$2" "$3" "$4" "$5" "$6" "$7" "$8" "$9"
		do
			if ! test -z "$n"
			then
				$tester $n
				if test "$?" = 0
				then
					db_add "$array" "$n" DB_fstools
					echo "export with_fs_${array}=\"${n}\"" >> Install-fstools.sh
					export $name="$n"
					echo "ok. (\"$n\", detected)"
					echo "[FS] (\"$n\", detected)" >> $LOG
					case $n in
						*unsafe*)
							echo ""
							echo "************************ WARNING **************************"
							echo "The only working $1 is $n, which is unsafe. Please check $n and whether you trust it."
							echo ""
							echo "Do you want to continue? (please type \"yes\" if you are sure)"
							echo ""
							read cont
							if ! test "$cont" = "yes"
							then
								echo "Critical: user didn't accept $n for $1."
								echo "[FS] Critical: user didn't accept $n for $1." >> $LOG
								exit 1
							fi
						;;
					esac							
					return
				fi
			fi
		done
		echo "FAILED."
		echo ""
		echo "Critical: Can't find a working $1."
		echo "[FS] Critical: Can't find a working $1." >> $LOG
		exit 1
	else
		$3 ${!2}
		if test "$?" = 0
		then
			db_add "$1" "${!2}" DB_fstools
			echo "export with_fs_${1}=\"${!2}\"" >> Install-fstools.sh
			echo "ok. (\"${!2}\", user supplied)"
			echo "[FS] \"${!2}\", user supplied" >> $LOG
		else
			echo "FAILED."
			echo ""
			echo "Critical: User supplied $1 (\"${!2}\") failed."
			echo "[FS] Critical: User supplied $1 (\"${!2}\") failed." >> $LOG
			exit 1
		fi
	fi

}

function fs_detect() {
	# IMPORTANT: order of these tests DOES matter!
	echo "[FS] fs_detect() start" >> $LOG

	fs_test	"tempfile" "with_fs_tempfile" fstest_tempfile "tempfile" "lib/tempfile.sh" "lib/tempfile_unsafe.sh"
	fs_test "cat" "with_fs_cat" fstest_cat "cat" "type"
	fs_test "head" "with_fs_head" fstest_head "head" "lib/head.sh"
	fs_test "mkdir" "with_fs_mkdir" fstest_mkdir "mkdir -p"
	fs_test "cp" "with_fs_cp" fstest_copy "cp -rpH" "cp -r"
	fs_test "mv" "with_fs_mv" fstest_move "mv"
	fs_test "rm" "with_fs_rm" fstest_remove "rm -rf"
	fs_test "ln" "with_fs_ln" fstest_link "ln -s" "ln" "cp"
	fs_test "awk" "with_fs_awk" test_awk "awk" "gawk" "mawk"
	fs_test "find" "with_fs_find" fstest_find "find" "/usr/bin/find" "lib/find.sh"
	fs_test "ar" "with_fs_ar" fstest_ar "${with_target}ar" "ar" "/usr/bin/ar"
	fs_test "ranlib" "with_fs_ranlib" fstest_ranlib "${with_target}ranlib" "ranlib" "/usr/bin/ranlib"

	echo "[FS] fs_detect() finish" >> $LOG
}

#----------------------------- Detect compiler tools -------------------------------

# Detect dlc trying a speficif dir
function dlc_detect_() {
	# For DLC target, check if we can find DLC
	cfg=`lib/dlc-config.sh $1 | $with_fs_head -n 1`
	if test -z "$cfg"
	then
		db_add sys_dlc_ld "" DB_osdep
		return 1
	fi

	dlc_detected="1"
	dlc_includes=`echo "$cfg" | awk -v "FS=|" '{print $1}'`
	dlc_libs=`echo "$cfg" | awk -v "FS=|" '{print $2}'`
	dlc_prefix=`echo "$cfg" | awk -v "FS=|" '{print $3}'`
	if ! test -z "$dlc_prefix"
	then
		dlc_prefix="${dlc_prefix}/bin/"
	fi

	with_sys_cflags="${dlc_includes} ${with_sys_cflags}"
	with_sys_ldl="${dlc_libs}"

	db_add sys_cflags "$with_sys_cflags" DB_osdep
	db_add sys_ldl "$with_sys_ldl" DB_osdep
	db_add sys_dlc_ld "${dlc_prefix}dlc-ld" DB_osdep

	dlc_ld="${dlc_prefix}dlc-ld"
	export dlc_ld
	export dlc_detected

	return 0;

}

# Detect dlc trying a list of dirs
function dlc_detect()
{
	for n in "$@"
	do
		dlc_detect_ $n && return 0
	done
	return 1
}

function compiler_detect() {
	if ! test -z "$with_target_dir"
	then
		echo "Forced target system: $with_target_dir"
		echo "Forced target system: $with_target_dir" >> $LOG
	fi
	if ! test -z "$with_host_dir"
	then
		echo "Forced host system: $with_host_dir"
		echo "Forced host system: $with_host_dir" >> $LOG
	fi

	echo -n "Detecting c compiler... "

	echo "[CC] Detecting c compiler... " >> $LOG

	CFLAGS="$with_cc_O $with_cc_debug"
	if test -z "$with_cc"
	then
		detect_compiler
		if test "$?" != 0
		then
			echo "FAILED."
			echo "Critical: can't find a working c compiler"
			echo "[CC] Critical: can't find a working c compiler" >> $LOG
			exit 1
		fi
		export with_cc=$CC
		db_add sys_cc "$with_cc" DB_osdep
		echo "ok. (\"$CC $CFLAGS\", detected)"
		echo "[CC] (\"$CC $CFLAGS\", detected)"	>> $LOG

	else
		CC=$with_cc
		check_compiler
		if test "$?" != 0
		then
			echo "Critical: c compiler failed (\"$with_cc $CFLAGS\", user supplied)"
			echo "[CC] Critical: c compiler failed (\"$with_cc $CFLAGS\", user supplied)"	>> $LOG
			exit 1
		fi
		echo "ok. (\"$CC $CFLAGS\", user supplied)"
		echo "[CC] ok. (\"$CC $CFLAGS\", user supplied)"	>> $LOG
	fi

	if test -z "$with_cc_host"
	then
		if ! test -z "$with_host_dir"
		then
			export with_cc_host="${with_host}gcc"
		else
			export with_cc_host=$with_cc
		fi
	fi
	db_add sys_cc_host "$with_cc_host" DB_osdep

	export with_cc CC CFLAGS

	if test "$with_os" = "windlc"
	then
		if test -z "$with_dlc_prefix"
		then
			echo -n "Detecting DL-Compat... "
			echo "[DLC] Detecting DL-Compat... " >> $LOG
			dlc_detect "/usr/local" "/usr"
		else
			echo -n "Testing DL-Compat... "
			echo "[DLC] Testing DL-Compat ($with_dlc_prefix)... " >> $LOG
			dlc_detect "$with_dlc_prefix"
		fi

		if test -z "$dlc_detected"
		then
			echo "[DLC] FAILED." >> $LOG
			echo "FAILED."
			echo ""
			echo "DL-Compat couldn't be found. Please install it before compiling GPMI."
			echo "DL-Compat can be found at http://www.libgpmi.org."
			exit 1
		else
			if test -z "${dlc_prefix}"
			then
				echo "found (global)"
				echo "[DLC] found global" >> $LOG
			else
				dlc_dirname=`dirname $dlc_prefix`
				echo "found ($dlc_dirname)"
				echo "[DLC] found $dlc_dirname" >> $LOG
			fi
		fi
	fi
}

#----------------------------- Detect scripts tools -------------------------------
function test_script() {
	tmp="with_$1"
	tmpp1="with_$1_params_libs"
	tmpp2="with_$2_params_includes"
	if test "${!tmp}" = "1"
	then
		if test -z "${!tmpp1}${!tmpp2}"
		then
			echo -n "Scripts: trying to configure $1... "
			echo "[SCRIPT] Scripts: trying to configure $1... "	>> $LOG
			if test -f "lib/$1-config.sh"
			then
				cfg=`lib/$1-config.sh | $with_fs_head -n 1`
				if test -z "$cfg"
				then
					echo "FAILED: not found. Deselected."
					echo "[SCRIPT] not found. Deselected." >> $LOG
					export with_$1=0
				else
					echo "ok (params: $cfg)"
					echo "[SCRIPT] ok (params: $cfg)" >> $LOG
					includes=`echo "$cfg" | awk -v "FS=|" '{print $1}'`
					libs=`echo "$cfg" | awk -v "FS=|" '{print $2}'`
					db_add "$1" "$includes" DB_includes
					db_add "$1" "$libs" DB_libs
				fi
			else
				echo "FAILED. (Can't find the config script). Deselected."
				echo "[SCRIPT] FAILED. (Can't find the config script). Deselected." >> $LOG
				export with_$1=0
			fi
		else
			echo -n "Scripts: using user supplied configuration for $1... "
			echo -n "[SCRIPT] using user supplied configuration for $1: '$tmpp1' '$tmpp2' " >> $LOG
			db_add "$1" "${!tmpp1}" DB_libs
			db_add "$1" "${!tmpp2}" DB_includes
			echo "Done."
		fi
	else
		echo "Scripts: $1 is not selected, skipping. "
		echo "[SCRIPTS] $1 is not selected, skipping. " >> $LOG
	fi
}

function script_detect() {
	echo "[SCRIPT] script_detect() start" >> $LOG
	for n in $scripts
	do
		test_script $n
	done
	echo "[SCRIPT] script_detect() finish" >> $LOG
}

#----------------------------- Create makefiles -------------------------------
function makefile_create() {
	echo "[MAKE] makefile_create start" >> $LOG

	# NOTE: I use the following formula for each generation:
	#   "cat input_file | subst DB | subst DB | ... | subst DB > output_file"
	# Instead of the initial cat, I could redirect the input file into the
	# first subst, but I believe this way it is all more readable.

	echo -n "Makefiles: creating makefiles for core packages... "
	echo "[MAKE] creating makefiles for core packages... " >> $LOG
	for n in `awk '/PACKAGES.*=/ {print $3}' < $src_root/packages/Makefile.am`
	do
		$with_fs_cat data/Makefile.core-pkg | subst DB_fstools > $src_root/packages/$n/Makefile
		mkdir -p $src_root/packages/$n/generated
	done
	echo "Done."

	with_example=1

	echo -n "Makefiles: creating makefiles for core modules... "
	echo "[MAKE] creating makefiles for core modules... " >> $LOG

	export DB_includes
	export DB_libs

	for n in `awk '/MODULES.*=/ {print $3}' < $src_root/modules/Makefile.am`
	do
		flag=with_$n
		if test "${!flag}" = "1"
		then
			tmp=""
			a1=`db_get "$n" DB_includes`
			db_add modcflags "$a1" tmp
			a2=`db_get "$n" DB_libs`
			db_add modldflags "$a2" tmp
			$with_fs_cat data/Makefile.core-mod  | subst tmp | subst DB_fstools > $src_root/modules/$n/Makefile
		else
			cp data/Makefile.empty $src_root/modules/$n/Makefile
		fi
	done
	echo "Done."

	echo -n "Makefiles: creating makefiles for example apps... "
	echo "[MAKE] creating makefiles for example apps... " >> $LOG
	for n in `awk '/EXAMPLES_gen.*=/ {print $3}' < $src_root/example-app/Makefile.am`
	do
		$with_fs_cat data/Makefile.examples | subst DB_fstools > $src_root/example-app/$n/Makefile
	done
	echo "Done."

	echo -n "Makefiles: creating Makefile.config... "
	echo "[MAKE] creating Makefile.config... " >> $LOG
	$with_fs_cat data/Makefile.config | subst DB_fstools | subst DB_osdep | subst DB_install_paths > $src_root/Makefile.config
	echo "Done."

	echo "[MAKE] makefile_create finish" >> $LOG
}

#--------------------- Create script test's shell script ---------------------------

function test_sh_create() {
	echo "[TEST] test_sh_create start" >> $LOG

$with_fs_cp data/test.sh $src_root/tests/app-scripts/test.sh
	echo "[TEST]  scripts: '$scripts' " >> $LOG
for n in $scripts
do
	tmp=with_$n
	if test "${!tmp}" = "1"
	then
	echo "[TEST]  generating test.sh for: '$n' " >> $LOG
	msg=`echo "Testing $n-module... " | awk '{tmp=$0; for(n=length(tmp); n<40; n++) { tmp=tmp " "} print tmp }'`
	echo "echo -n \"$msg\"" >> $src_root/tests/app-scripts/test.sh

	test_file="script"

	# XXX: this should be replaced; this info should be data, not code.
	# cli has another output because forking
	[ $n = "cli" ] && test_file="cli"

	echo "[TEST]  test_file: '$test_file' " >> $LOG

	echo '

# make sure we will not use an old out
'$with_fs_rm' out
$EMULATOR $EXE '$n' example_'$n' > out 2> /dev/null
../test.awk '$test_file'.OK < out || fail=1
'$with_fs_rm' out

	' >> $src_root/tests/app-scripts/test.sh
	else
		echo "# Not testing $n as it is deselected" >> $src_root/tests/app-scripts/test.sh
	fi
done

	echo '
if ! test -z "$fail"
then
	echo "FAILED (one or more script module(s) has/have failed)"
	false
else
	true
fi
	'>> $src_root/tests/app-scripts/test.sh

	echo "[TEST] test_sh_create finish" >> $LOG
}

function install_create() {
	echo "[INSTALL] install_create start" >> $LOG

	echo -n "Creating extra install variables... "
	echo "export with_os=$with_os" >> Install-fstools.sh
	echo "export with_sys_libext_dynamic=$with_sys_libext_dynamic" >> Install-fstools.sh
	echo "export with_sys_libext_exe=$with_sys_libext_exe" >> Install-fstools.sh
	echo "" >> Install-fstools.sh
	echo "# Installation paths " >> Install-fstools.sh
	echo "$DB_install_paths "| awk -v "FS=>" '{print "export gpmi_" $1 "=" $2}' >> Install-fstools.sh
	echo "Done."

	echo -n "Creating gpmi-config... "
	tmp=""
	db_add vars "$with_install_path_lib_gpmi/Install-fstools.sh" tmp
	db_add with_fs_awk "$with_fs_awk" tmp
	db_add with_fs_rm "$with_fs_rm" tmp
	db_add with_fs_cp "$with_fs_cp" tmp

	db_add listgen "host.listgen${with_sys_libext_exe}" tmp
	db_add wrapgen "host.wrapgen${with_sys_libext_exe}" tmp
	if ! test -z "$dlc_includes"
	then
		db_add includes "${dlc_includes}" tmp
	else
		db_add includes "" tmp
	fi

	# Derive the sed script that can replace ### fields of fstools and install_paths 
	replace=`echo "$DB_install_paths
$DB_fstools" | $with_fs_awk -v "FS=>" '
	{
		if (/^install_path/) {$1="gpmi_" $1} else {$1="with_fs_" $1}
		printf("	s@###%s###@$%s@g;###newline###", $1, $1)
	}
	'`

	db_add replace_install_paths "$replace" tmp
	subst tmp < data/gpmi-config | sed "s/###newline###/\n/g" > gpmi-config
	chmod 0755 gpmi-config
	echo "Done."
	echo "[INSTALL] install_create finish" >> $LOG
}

function makefile_vars() {
	with_sys_cflags="${with_cc_O}   ${with_cc_debug}   ${with_sys_cflags}"
	db_add sys_cflags "$with_sys_cflags" DB_osdep
	db_add sys_ldflags "$with_sys_ldflags" DB_osdep
}

function set_options_() {
	local target_var
	local target_opt

	target_var="with_sys_$opt"
	target_opt="options_$opt"
	echo "[OPTIONS] setting $target_var; before: ${!target_var}" >> $LOG
	for n in ${!target_opt}
	do
		name=${n%%>*}
		value=${n#*>}
		echo "[OPTIONS]   $name (${!name}) = $value" >> $LOG
		if test "${!name}" = 1
		then
			export $target_var="${!target_var} $value"
		fi
	done
	echo "[OPTIONS] setting $target_var;  after: ${!target_var}" >> $LOG
}

function set_options() {
	echo -n "Setting up options... "

	# Include options here so variables are evaluated
	. data/configure_options.tbl

	for opt in $option_tables
	do
		echo -n "[$opt] "
		set_options_
	done

	echo "done."
}




#----------------------------- Main -----------------------------------------

export LOG
echo -n "[START]" > $LOG
date >> $LOG

if test -f "local.conf"
then
	echo "[CONF] Found local conf in local.conf, including" >> $LOG
	. local.conf
fi

echo "[cmdline] Parsing command line arguments..." >> $LOG
# Set up base configuration
process_params "$@"



echo "Configuring gpmi..."
echo ""

echo "">Install-fstools.sh

set_install_paths


# Detect operating system
os_detect

# Detect fs tools
fs_detect

# Detect compiler
compiler_detect

# Detect libs
if test "$with_option_threads" == "1"
then
	. lib/pthreads-config.sh
fi

# Set options
set_options

# Detect script modules
script_detect

# Set up makefile variables
makefile_vars

# Create makefiles
makefile_create

# Create test.sh for scripts
test_sh_create

# Create extra install variables
install_create
