/*
    libgpmi - General Package/Module Interface
    Copyright (C) 2005-2007 Patric 'TrueLight' Stout & Tibor 'Igor2' Palinkas
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "error.h"
#include "threads.h"

const char *gpmi_error_sources_txt[] = {
	"Unknown",
	"gpmi_kw_add()",
	"gpmi_kw_add_auto",
	"gpmi_kw_find_multi",
	"gpmi_kw_del",
	"gpmi_libdep_load",
	"gpmi_find_path",
	"gpmi_load_library",
	"gpmi_load_module",
	"gpmi_pkg_load",
	"gpmi_library_load_lowlevel",
	NULL
};

const char *gpmi_error_txt[] = {
	NULL,
	"Failed to load library",	   /* gpmi_err_load_library */
	"Unexpected null string",    /* gpmi_err_null_str */
	"Version mismatch",          /* gpmi_err_version */
	"Initialization failed",     /* gpmi_err_init_failed */
	"String too long",           /* gpmi_err_too_long */
	"Already loading library",   /* gpmi_err_already_loading */
	"Found cyclic dependency",   /* gpmi_err_cyclic_dep */
	"Keyword already exist",     /* gpmi_err_kw_already_exists */
	"Keyword table not autoinc", /* gpmi_err_kw_not_autoinc */
	"Keyword table is not int",  /* gpmi_err_kw_not_int */
	"Keyword not found",         /* gpmi_err_kw_not_found */
	"File not found",            /* gpmi_err_file_not_found */
	"No regcmd was found",       /* gpmi_err_init_not_found */
	"Entry point not found",     /* gpmi_err_entry_not_found */
};

#define array_len(array) (sizeof(array) / sizeof(char *))

const char *gpmi_strerror(int gpmi_errno)
{
	/* Check if it is a gpmi error */
	if ((gpmi_errno > -34000) || (gpmi_errno < -34000 - array_len(gpmi_error_txt)))
		return NULL;

	return gpmi_error_txt[ (-gpmi_errno) - 34000];
}



const char *gpmi_strerror_src(int srcno)
{
	/* Check if it is a gpmi error */
	if ((srcno < 0) || (srcno >= array_len(gpmi_error_sources_txt)))
		return NULL;

	return gpmi_error_sources_txt[srcno];
}

void gpmi_err_stack_push(gpmi_error_t error, gpmi_error_source_t from, const char *details)
{
	gpmi_err_stack_t *entry;
	gpmi_private_globals_t *priv;

	/* Don't push anything if the stack is not enabled */
	if (!gpmi_shared.error_stack_enable)
		return;

	/* Create the entry */
	entry = malloc(sizeof(gpmi_err_stack_t));
	entry->error = error;
	entry->from  = from;
	if (details != NULL)
		entry->details = strdup(details);
	else
		entry->details = NULL;

	/* Access the private area - we need our error stack */
	priv = gpmi_threads_get_private();

	/* Insert at the beginning of the linked list */
	entry->next          = priv->err_stack;
	priv->err_stack = entry;
}

gpmi_err_stack_t *gpmi_err_stack_pop(void)
{
	gpmi_err_stack_t *entry;
	gpmi_private_globals_t *priv;

	/* Access the private area - we need our error stack */
	priv = gpmi_threads_get_private();

	/* Get the first record */
	entry = priv->err_stack;

	/* Set next record so we popped the first */
	if (entry != NULL)
		priv->err_stack = entry->next;

	return entry;
}

void gpmi_err_stack_destroy(gpmi_err_stack_t *entry)
{
	gpmi_err_stack_t *e, *next;
	gpmi_private_globals_t *priv;

	/* Access the private area - we need our error stack */
	priv = gpmi_threads_get_private();

	if (entry == NULL) {
		/* Destroy the whole stack */
		for (e = priv->err_stack; e != NULL; e = next) {
			next = e->next;
			gpmi_err_stack_destroy(e);
		}
		priv->err_stack = NULL;
	}
	else {
		/* Destroy one entry */
		if (entry->details != NULL)
			free(entry->details);

		free(entry);
	}
}

char *gpmi_str_err_stack_entry(gpmi_err_stack_t *entry)
{
	char *tmp, *det;
	const char *from, *msg;

	msg = gpmi_strerror(entry->error);
	if (msg == NULL)
		msg = "(unknown)";

	from = gpmi_strerror_src(entry->from);
	if (from == NULL)
		from = "(unknown)";

	if (entry->details != NULL)
		det = entry->details;
	else
		det = "(empty)";

	tmp = malloc(strlen(msg) + strlen(from) + strlen(det) + 32);
	sprintf(tmp, "%s(): '%s'; details: '%s'", from, msg, det);
	return tmp;
}

void gpmi_err_stack_process_str(void (*callback)(gpmi_err_stack_t *entry, char *string))
{
	gpmi_err_stack_t *entry;
	char *str;

	assert(callback != NULL);
	do {
		entry = gpmi_err_stack_pop();
		if (entry != NULL) {
			str   = gpmi_str_err_stack_entry(entry);
			callback(entry, str);
			free(str);
			gpmi_err_stack_destroy(entry);
		}
	} while(entry != NULL);
}

int gpmi_err_stack_enable(void)
{
	int previous;

	gpmi_shared_lock;
	previous = gpmi_shared.error_stack_enable;
	gpmi_shared.error_stack_enable = 1;
	gpmi_shared_unlock;
	return previous;
}

int gpmi_err_stack_disable(void)
{
	int previous;

	gpmi_shared_lock;
	previous = gpmi_shared.error_stack_enable;
	gpmi_shared.error_stack_enable = 0;
	gpmi_shared_unlock;
	return previous;
}
