/*
    libgpmi - General Package/Module Interface
    Copyright (C) 2005-2007 Patric 'TrueLight' Stout & Tibor 'Igor2' Palinkas
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/**
 * @file events.h
 * \brief Handles the event-calls in GPMI.
 *
 * Handles creating, binding, calling and removing events within GPMI and handles
 *  all the calls to the modules and packages.
 */

#ifndef GPMI_EVENTS_H
#define GPMI_EVENTS_H

#include <gpmi/keywords.h>
#include <gpmi/libs.h>
#include <gpmi/libs-module.h>

int gpmi_event_register(const char *event_name, const char *param_types);
int gpmi_event_find(const char *event_name, char **param_types);
const char *gpmi_event_get_types(int event_num);

/* 3rd param of those two function should be void *, not gpmi_ProcEvent, while it is one, but
 *  because we fill in params on the ..., the compiler things it is wrong ;)
 */
gpmi_global_event_t gpmi_event_bind_global(gpmi_module *mod, const void *handler);
gpmi_ProcEvent *gpmi_event_bind(gpmi_module *mod, int event, const void *handler);

void gpmi_event_init(void);
void gpmi_event_uninit(void);
void gpmi_event_close(gpmi_module *mod);
void gpmi_event_open(gpmi_module *mod);



/* Identical to wrapgen's types */
typedef enum gpmi_globev_types {
	gpmi_globev_type_keyword = 0, /* -> integer */
	gpmi_globev_type_int     = 1, /* -> integer */
	gpmi_globev_type_string  = 2, /* -> string */
	gpmi_globev_type_double  = 3, /* -> dbl */
	gpmi_globev_type_void    = 4, /* -> NONE (not used) */
	gpmi_globev_type_uint    = 5, /* -> uinteger */
	gpmi_globev_type_ptr     = 6, /* -> ptr */
	gpmi_globev_type_uchar   = 7, /* -> uchar */
} gpmi_globev_type_t;

struct gpmi_globev_param_t {
	gpmi_globev_type_t type;
	union {
		void *ptr;
		char *string;
		int integer;
		unsigned int uinteger;
		double dbl;
		unsigned char uchar;
	} data;
};

/* Global event handler: */
void gpmi_global_event_parse(gpmi_module *module, int event_id, ...);


#ifndef NO_VARARG_MACRO
/**
 * \brief The macro to call an event of a module.
 *
 * Example on calling: gpmi_event(mod, id, p1, p2);
 * Negative event-ids are ignored.
 */
#define gpmi_event(gpmi_module, event, parameters...) \
{	\
	if ((event) >= 0 && (gpmi_module)->proc_EventGlobal != NULL) { \
		gpmi_set_current_module(gpmi_module); \
		gpmi_global_event_parse((gpmi_module), (event), parameters); \
	} \
	if ((event) >= 0 && (event) < (gpmi_module)->event_next_free) { \
		if ((gpmi_module)->proc_EventArray[(event)] != NULL) { \
			gpmi_set_current_module(gpmi_module); \
			(gpmi_module)->proc_EventArray[(event)]((gpmi_module), (event), parameters); \
		} \
	} \
}

#else /* NO_VARARG_MACRO */
void gpmi_event(gpmi_module *module, int event, ...);
#endif /* NO_VARARG_MACRO */

#endif /* GPMI_EVENTS_H */
