/*
    libgpmi - General Package/Module Interface
    Copyright (C) 2005-2007 Patric 'TrueLight' Stout & Tibor 'Igor2' Palinkas
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "libs.h"
#include "libs-package.h"
#include "os_dep.h"
#include "error.h"
#include "global.h"
#include "threads.h"

/**
 * Load a named package, register commands with the provided registering function.
 *
 * @param required_version The version the package should have. 0 to disable check.
 */
int gpmi_pkg_load(const char *library_name, int required_version, gpmi_PkgProcRegisterCommand *regcmd, void *handle, gpmi_package **pkg_handle)
{
	gpmi_library *library;
	int state;

	/* If we fail, leave NULL in the handle so later calls using handle will fail too */
	if (pkg_handle != NULL)
		*pkg_handle = NULL;

	if (library_name == NULL) {
		gpmi_err_stack_push(gpmi_err_null_str, gpmi_err_src_pkg_load, NULL);
		return gpmi_err_null_str;
	}

	/* Load the lib, make a package */
	library = gpmi_load_library(gpmi_shared.path_packages, library_name, &state, 1);
	if (library == NULL) {
		if (state < 0)
			return state;

		gpmi_err_stack_push(gpmi_err_load_library, gpmi_err_src_pkg_load, library_name);
		return gpmi_err_load_library;
	}
	return gpmi_pkg_load_lib(library, required_version, regcmd, handle, pkg_handle);
}

/**
 * Load a package from an already loaded lib; register commands with the
 * provided registering function.
 *
 * @param required_version The version the package should have. 0 to disable check.
 */
int gpmi_pkg_load_lib(gpmi_library *library, int required_version, gpmi_PkgProcRegisterCommand *regcmd, void *handle, gpmi_package **pkg_handle)
{
	gpmi_package *package;
	char *library_name;

#ifndef gpmi_no_static
	if ((library->flags & gpmi_libf_static) == gpmi_libf_static)
		/* Use the pointer in the library for static packages */
		package = library->static_package;
	else
#endif
	/* Allocate new package structure for dynamic packages */
	package = gpmi_pkg_alloc(library);

	/* Check version, if we need to */
	if (package->proc_CheckVer != NULL && required_version != 0 && package->proc_CheckVer(required_version) != 0) {
		/* Version check failed, we unload and quit */
		library_name = strdup(library->name);
		gpmi_pkg_unload(package);
		gpmi_err_stack_push(gpmi_err_version, gpmi_err_src_pkg_load, library_name);
		free(library_name);
		return gpmi_err_version;
	}

	/* For first load, we run proc_Init, if it exists and the package is not static */
	if ((library->references == 1) && (package->proc_Init != NULL) &&
			((library->flags & gpmi_libf_static) == 0) && (package->proc_Init() != 0)) {
		/* Init failed, we unload and quit */
		library_name = strdup(library->name);
		gpmi_pkg_unload(package);
		gpmi_err_stack_push(gpmi_err_init_failed, gpmi_err_src_pkg_load, library_name);
		free(library_name);
		return gpmi_err_init_failed;
	}

	/* Check if we need to register commands for the caller */
	if (regcmd != NULL) {
		/* We need to register commands, and the package can't do it -> error */
		if (package->proc_RegCmd == NULL) {
			library_name = strdup(library->name);
			gpmi_pkg_unload(package);
			gpmi_err_stack_push(gpmi_err_init_not_found, gpmi_err_src_pkg_load, library_name);
			free(library_name);
			return gpmi_err_init_not_found;
		}

		package->proc_RegCmd(regcmd, handle, package);
	}

	if (pkg_handle == NULL) {
		/* The user doesn't want to handle the package, so we save it in a linked
		 *  list, and free it at gpmi_uninit(). Just a little bit of
		 *  service for our users ;)
		 */
		gpmi_shared_lock;
		package->next = gpmi_shared.dynamic_packages;
		gpmi_shared.dynamic_packages = package;
		gpmi_shared_unlock;
	}
	else {
		/* Set handle to package - we do not keep track of these packages */
		*pkg_handle = package;
	}

	return 0;
}

/**
 * Allocate a package structure
 */
gpmi_package *gpmi_pkg_alloc(gpmi_library *library)
{
	gpmi_package *package;
	int ns_len;

	package = malloc(sizeof(gpmi_package));
	memset(package, 0, sizeof(package));

	package->library = library;

	/* First get the name space */
	package->name_space = gpmi_dlsym_lowlevel(package->library, PREFIX_GPMI_PKG_GEN "dynamic_name_space");
	assert(package->name_space != NULL);
	ns_len = strlen(*package->name_space);

	package->proc_Init     = gpmi_dlsym_lowlevel(package->library, gpmi_pkg_misc_func(*package->name_space, ns_len, "init", 4));
	package->proc_UnInit   = gpmi_dlsym_lowlevel(package->library, gpmi_pkg_misc_func(*package->name_space, ns_len, "uninit", 6));
	package->proc_CheckVer = gpmi_dlsym_lowlevel(package->library, gpmi_pkg_misc_func(*package->name_space, ns_len, "checkver", 8));
	package->proc_RegCmd   = gpmi_dlsym_lowlevel(package->library, gpmi_pkg_misc_func(*package->name_space, ns_len, "register_commands", 17));

	return package;
}

/**
 * Free each dynamic field of a package structure and the pointer itself as well.
 */
void gpmi_pkg_free(gpmi_package *package)
{
	/* No dynamic fields yet. This routine is called at least from 3 places
	   so please do not remove it as an optimization!
	*/
	free(package);
}

/**
 * Load a package, register commands with the provided registering function
 */
void gpmi_pkg_unload(gpmi_package *package)
{
	int tmp;

	/* If we have a library, free it */
	if (package->library != NULL) {
		tmp = package->library->flags;

		/* Unload the library */
		gpmi_unload_library(package->library, package->proc_UnInit);

		/* Totally free the package */
#ifndef gpmi_no_static
		if ((tmp & gpmi_libf_static) == 0)
#endif
			gpmi_pkg_free(package);
	} else {
		/* Free the package */
		gpmi_pkg_free(package);
	}
}

/**
 * Resolve a symbol in a package.
 */
void *gpmi_pkg_resolve(gpmi_package *package, const char *name)
{
	return gpmi_dlsym_lowlevel(package->library, name);
}

/**
 * Assamble the name of an entry point in a dynamic package
 */
char *gpmi_pkg_misc_func(char *name_space, int name_space_len, char *func, int func_len)
{
	int len;
	gpmi_private_globals_t *priv;

	priv = gpmi_threads_get_private();

	/* Realloc space if needed */
	len = name_space_len + func_len +	PREFIX_GPMI_PKG_GEN_LEN + 8;
	if (priv->pkg_misc_func_size < len) {
		len += 32;
		if (priv->pkg_misc_func_buff != NULL)
			free(priv->pkg_misc_func_buff);

		priv->pkg_misc_func_buff = malloc(len);
		priv->pkg_misc_func_size = len;
		strcpy(priv->pkg_misc_func_buff, PREFIX_GPMI_PKG_GEN);
	}

	/* We add the same things many time. A good c compiler should optimize it */

	/* Copy strings together - PEFIX_GPMI_PKG_GEN is already there */
	strcpy(priv->pkg_misc_func_buff + PREFIX_GPMI_PKG_GEN_LEN, name_space);
	strcpy(priv->pkg_misc_func_buff + PREFIX_GPMI_PKG_GEN_LEN + name_space_len + 1, func);

	/* Add separator and terminator */
	priv->pkg_misc_func_buff[PREFIX_GPMI_PKG_GEN_LEN + name_space_len] = '_';
	priv->pkg_misc_func_buff[PREFIX_GPMI_PKG_GEN_LEN + name_space_len + func_len + 1] = '\0';

	return priv->pkg_misc_func_buff;
}
